import fsspec

from rubicon_ml.client import Artifact as SyncArtifact


class Artifact(SyncArtifact):
    """An asynchronous client artifact.

    An `artifact` is a catch-all for any other type of
    data that can be logged to a file.

    For example, a snapshot of a trained model (.pkl) can
    be logged to the `experiment` created during its run.
    Or, a base model for the model in development can be
    logged to a `project` when leveraging transfer learning.

    An `artifact` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Artifact
        The artifact domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    async def _get_data(self):
        """Overrides `rubicon.client.Artifact._get_data` to
        asynchronously load the data associated with this artifact.
        """
        project_name, experiment_id = self.parent._get_parent_identifiers()

        self._data = await self.repository.get_artifact_data(
            project_name, self.id, experiment_id=experiment_id
        )

    async def download(self, location=None, name=None):
        """Overrides `rubicon.client.Artifact.download` to
        asynchronously download this artifact's data.

        Parameters
        ----------
        location : str, optional
            The S3 bucket to download the artifact to. S3 buckets
            must be prepended with 's3://'.
        name : str, optional
            The name to give the downloaded artifact file.
            Defaults to the artifact's given name when logged.
        """
        if location is None:
            location = "."

        if name is None:
            name = self._domain.name

        with fsspec.open(f"{location.rstrip('/')}/{name}", "wb", auto_mkdir=False) as f:
            f.write(await self.data)

    @property
    async def data(self):
        """Overrides `rubicon.client.Artifact.data` to
        asynchronously get the artifact's raw data.
        """
        if self._data is None:
            await self._get_data()

        return self._data



from rubicon_ml.client import Dataframe as SyncDataframe
from rubicon_ml.client.asynchronous import TagMixin


class Dataframe(TagMixin, SyncDataframe):
    """An asynchronous client dataframe.

    A `dataframe` is a two-dimensional, tabular dataset with
    labeled axes (rows and columns) that provides value to the
    model developer and/or reviewer when visualized.

    For example, confusion matrices, feature importance tables
    and marginal residuals can all be logged as a `dataframe`.

    A `dataframe` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Dataframe
        The dataframe domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    pass



from rubicon_ml.client.base import Base  # noqa F401
from rubicon_ml.client.config import Config

from rubicon_ml.client.mixin import ArtifactMixin, DataframeMixin, TagMixin  # noqa F401

from rubicon_ml.client.artifact import Artifact
from rubicon_ml.client.dataframe import Dataframe
from rubicon_ml.client.feature import Feature
from rubicon_ml.client.metric import Metric
from rubicon_ml.client.parameter import Parameter
from rubicon_ml.client.experiment import Experiment
from rubicon_ml.client.project import Project
from rubicon_ml.client.rubicon import Rubicon

__all__ = [
    "Artifact",
    "Config",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "Rubicon",
]



import warnings

from rubicon_ml.client import Base, TagMixin
from rubicon_ml.exceptions import RubiconException


class Dataframe(Base, TagMixin):
    """A client dataframe.

    A `dataframe` is a two-dimensional, tabular dataset with
    labeled axes (rows and columns) that provides value to the
    model developer and/or reviewer when visualized.

    For example, confusion matrices, feature importance tables
    and marginal residuals can all be logged as a `dataframe`.

    A `dataframe` is logged to a `project` or an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Dataframe
        The dataframe domain model.
    parent : rubicon.client.Project or rubicon.client.Experiment
        The project or experiment that the artifact is
        logged to.
    """

    def __init__(self, domain, parent):
        super().__init__(domain, parent._config)

        self._data = None
        self._parent = parent

    def get_data(self, df_type="pandas"):
        """Loads the data associated with this Dataframe
        into a `pandas` or `dask` dataframe.

        Parameters
        ----------
        df_type : str, optional
            The type of dataframe. Can be either `pandas` or `dask`.
        """
        project_name, experiment_id = self.parent._get_parent_identifiers()

        self._data = self.repository.get_dataframe_data(
            project_name,
            self.id,
            experiment_id=experiment_id,
            df_type=df_type,
        )

        return self._data

    def plot(self, df_type="pandas", **kwargs):
        """Render the dataframe using `hvplot`.

        Parameters
        ----------
        df_type : str, optional
            The type of dataframe. Can be either `pandas` or `dask`.
        kwargs : dict
            Additional keyword arguments to be passed along to the
            `hvplot` function.

        Notes
        -----
        For usage, visit: https://hvplot.holoviz.org/user_guide/Plotting.html
        For customizations, visit:
        https://hvplot.holoviz.org/user_guide/Customization.html

        Examples
        --------
        >>> # Log a line plot
        >>> dataframe.plot(kind='line', x='Year', y='Number of Subscriptions')
        """
        try:
            if df_type == "pandas":
                import hvplot.pandas  # noqa F401
            else:
                import hvplot.dask  # noqa F401
        except ImportError:
            raise RubiconException(
                "`hvplot` is required for plotting. Install with `pip install hvplot`."
            )

        return self.get_data(df_type=df_type).hvplot(**kwargs)

    @property
    def id(self):
        """Get the dataframe's id."""
        return self._domain.id

    @property
    def description(self):
        """Get the dataframe's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the time this dataframe was created."""
        return self._domain.created_at

    @property
    def data(self):
        """Get the dataframe's data as it was logged."""
        warnings.warn(
            "`data` is deprecated, use `get_data()` instead",
            DeprecationWarning,
        )
        if self._data is None:
            self.get_data()

        return self._data

    @property
    def parent(self):
        """Get the dataframe's parent client object."""
        return self._parent



import warnings

from rubicon_ml import __version__  # noqa: F401
from rubicon_ml import (
    Artifact,
    Dataframe,
    Experiment,
    Feature,
    Metric,
    Parameter,
    Project,
    Rubicon,
)

rename_warning_message = (
    "`rubicon` will be renamed to `rubicon_ml` in version 0.3.0. You "
    "can switch your imports now: `from rubicon_ml import Rubicon`."
)
warnings.warn(rename_warning_message)

__all__ = [
    "Artifact",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "Rubicon",
]



import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Input, Output

from rubicon_ml.ui.app import app


def make_project_selection_layout():
    """The html layout for the dashboard's project selection view."""

    return html.Div(
        id="project-selection",
        className="project-selection",
        children=[
            html.Div(
                className="project-selection--header",
                children=[
                    html.Div(className="project-selection--label", children="Projects"),
                    html.I(
                        id="project-selection--refresh-projects-btn",
                        n_clicks=0,
                        className="fas fa-sync-alt",
                    ),
                ],
            ),
            html.Div(id="project-selection--list", className="project-selection--list"),
        ],
    )


@app.callback(
    Output("project-selection--list", "children"),
    [Input("project-selection--refresh-projects-btn", "n_clicks")],
)
def _update_project_options(n_clicks):
    """Use the Rubicon client to load the available projects."""
    app._rubicon_model.update_projects()

    return dbc.ListGroup(
        children=[
            dbc.ListGroupItem(
                id={"type": "project-selection--project", "index": project["value"]},
                className="project-selection--project",
                children=project["value"],
                n_clicks=0,
                action=True,
            )
            for project in app._rubicon_model.project_options
        ]
    )



from rubicon_ml import __version__
from rubicon_ml.client import Rubicon
from rubicon_ml.intake_rubicon.base import DataSourceMixin


class ExperimentSource(DataSourceMixin):
    """An Intake data source for reading `rubicon` experiments.

    Parameters
    ----------
    urlpath : str
        The root directory the `rubicon` data is logged to.
    project_name : str
        The name of the `rubicon` project to load.
    experiment_id : str
        The ID of the `rubicon` experiment to load.
    """

    version = __version__

    container = "python"
    name = "rubicon_ml_experiment"

    def __init__(
        self, urlpath, project_name, experiment_id, metadata=None, storage_options=None, **kwargs
    ):
        self._experiment_id = experiment_id

        super().__init__(
            urlpath, project_name, metadata=metadata, storage_options=storage_options, **kwargs
        )

    def _get_schema(self):
        """Load the experiment with ID `self._experiment_id`
        from the project named `self._project_name`.
        """
        self._rubicon = Rubicon(persistence="filesystem", root_dir=self._urlpath)
        self._rubicon_object = self._rubicon.get_project(self._project_name).experiment(
            self._experiment_id
        )

        self._metadata.update(
            {
                "experiment": {
                    "name": self._rubicon_object.name,
                    "id": self._rubicon_object.id,
                    "created_id": self._rubicon_object.created_at,
                    "commit_hash": self._rubicon_object.commit_hash,
                    "tags": self._rubicon_object.tags,
                    "project": self._rubicon_object.project,
                }
            }
        )

        return super()._get_schema()



import click

from rubicon_ml.ui import Dashboard


@click.group()
@click.version_option()
def cli():
    pass


# Top level CLI commands


@cli.command(
    context_settings=dict(
        ignore_unknown_options=True,
    )
)
@click.option(
    "--root-dir",
    type=click.STRING,
    help="The absolute path to the top level folder holding the Rubicon project(s).",
    required=True,
)
@click.option(
    "--page-size",
    "-ps",
    type=click.IntRange(min=1),
    help="The number of rows that will be displayed on a page within the experiment table.",
    default=10,
)
@click.option(
    "--host", "-h", type=click.STRING, help="The host to serve the dashboard on.", default=None
)
@click.option(
    "--port", "-p", type=click.STRING, help="The port to serve the dashboard on.", default=None
)
@click.option(
    "--debug", "-d", type=click.BOOL, help="Whether or not to run in debug mode.", default=False
)
@click.argument("storage_options", nargs=-1, type=click.UNPROCESSED)
def ui(root_dir, host, port, debug, page_size, storage_options):
    """Launch the Rubicon Dashboard."""
    # convert the additional storage options into a dict
    # coming in as: ('--key1', 'one', '--key2', 'two')
    storage_options_dict = {
        storage_options[i][2:]: storage_options[i + 1] for i in range(0, len(storage_options), 2)
    }
    dashboard = Dashboard("filesystem", root_dir, page_size=page_size, **storage_options_dict)

    server_kwargs = dict(debug=debug, port=port, host=host)
    server_kwargs = {k: v for k, v in server_kwargs.items() if v is not None}
    dashboard.run_server(**server_kwargs)


# CLI groups

if __name__ == "__main__":
    cli()



from sklearn.pipeline import Pipeline

from rubicon_ml.sklearn.estimator_logger import EstimatorLogger


class RubiconPipeline(Pipeline):
    """An extension of `sklearn.pipeline.Pipeline` that automatically
    creates a Rubicon `experiment` under the provided `project` and logs
    the pipeline's `parameters` and `metrics` to it.

    A single pipeline run will result in a single `experiment` logged with
    its corresponding `parameters` and `metrics` pulled from the pipeline's
    estimators.

    Parameters
    ----------
    project : rubicon.client.Project
        The rubicon project to log to.
    steps : list
        List of (name, transform) tuples (implementing fit/transform) that are chained,
        in the order in which they are chained, with the last object an estimator.
    user_defined_loggers : dict, optional
        A dict mapping the estimator name to a corresponding user defined logger.
        See the example below for more details.
    experiment_kwargs : dict, optional
        Additional keyword arguments to be passed to
        `project.log_experiment()`.
    kwargs : dict
        Additional keyword arguments to be passed to
        `sklearn.pipeline.Pipeline()`.

    Example
    -------
    >>> pipeline = RubiconPipeline(
    >>>     project,
    >>>     [
    >>>         ("vect", CountVectorizer()),
    >>>         ("tfidf", TfidfTransformer()),
    >>>         ("clf", SGDClassifier()),
    >>>     ],
    >>>     user_defined_loggers = {
    >>>         "vect": FilterEstimatorLogger(select=["input", "decode_error", "max_df"]),
    >>>         "tfidf": FilterEstimatorLogger(ignore_all=True),
    >>>         "clf": FilterEstimatorLogger(ignore=["alpha", "penalty"]),
    >>>     }
    >>> )
    """

    def __init__(
        self,
        project,
        steps,
        user_defined_loggers={},
        experiment_kwargs={"name": "RubiconPipeline experiment"},
        **kwargs
    ):
        self.project = project
        self.user_defined_loggers = user_defined_loggers
        self.experiment_kwargs = experiment_kwargs

        self.experiment = None

        super().__init__(steps, **kwargs)

    def fit(self, X, y=None, tags=None, **fit_params):
        """Fit the model and automatically log the `fit_params`
        to Rubicon. Optionally, pass `tags` to update the experiment's
        tags.

        Parameters
        ----------
        X : iterable
            Training data. Must fulfill input requirements of first step of the pipeline.
        y : iterable, optional
            Training targets. Must fulfill label requirements for all steps of the pipeline.
        tags : list, optional
            Additional tags to add to the experiment during the fit.
        fit_params : dict
            Additional keyword arguments to be passed to
            `sklearn.pipeline.Pipeline.fit()`.
        """
        pipeline = super().fit(X, y, **fit_params)

        self.experiment = self.project.log_experiment(**self.experiment_kwargs)
        if tags is not None:
            self.experiment.add_tags(tags)

        for step_name, estimator in self.steps:
            logger = self.get_estimator_logger(step_name, estimator)
            logger.log_parameters()

        return pipeline

    def score(self, X, y=None, sample_weight=None):
        """Score with the final estimator and automatically
        log the results to Rubicon.

        Parameters
        ----------
        X : iterable
            Data to predict on. Must fulfill input requirements of first step of the pipeline.
        y : iterable, optional
            Targets used for scoring. Must fulfill label requirements for all steps of the pipeline.
        sample_weight : list, optional
            If not None, this argument is passed as sample_weight keyword argument to the
            score method of the final estimator.
        """
        score = super().score(X, y, sample_weight)

        logger = self.get_estimator_logger()
        logger.log_metric("score", score)

        return score

    def get_estimator_logger(self, step_name=None, estimator=None):
        """Get a logger for the estimator. By default, the logger will
        have the current experiment set.
        """
        logger = self.user_defined_loggers.get(step_name) or EstimatorLogger()

        logger.set_experiment(self.experiment)

        if step_name:
            logger.set_step_name(step_name)

        if estimator:
            logger.set_estimator(estimator)

        return logger



def _check_for_ui_extras():
    try:
        import dash  # noqa F401
        import dash_html_components as html  # noqa F401
    except ImportError:
        install_command = "pip install rubicon[ui]"
        message = f"Install the packages required for the UI with `{install_command}`."

        raise ImportError(message)


_check_for_ui_extras()

from rubicon_ml.ui import Dashboard  # noqa E402

__all__ = ["Dashboard"]



import warnings


class EstimatorLogger:
    """The base logger for sklearn estimators. By default,
    it will log all of the estimator's parameters.

    Parameters
    ----------
    estimator : a sklearn estimator, optional
        The estimator.
    experiment : rubicon.client.Experiment, optional
        The experiment to log the parameters and metrics to.
    step_name : str, optional
        The name of the pipeline step.
    """

    def __init__(self, estimator=None, experiment=None, step_name=None):
        self.estimator = estimator
        self.experiment = experiment
        self.step_name = step_name

    def _log_parameter_to_rubicon(self, name, value):
        try:
            self.experiment.log_parameter(name=f"{self.step_name}__{name}", value=value)
        except Exception:
            warning = (
                f"step '{self.step_name}' failed to write parameter '{name}' with value {value} "
                f"of type {type(value)}. try using the `FilterEstimatorLogger` with `ignore=['{name}']`"
            )

            warnings.warn(warning)

    def log_parameters(self):
        for name, value in self.estimator.get_params().items():
            self._log_parameter_to_rubicon(name, value)

    def log_metric(self, name, value):
        self.experiment.log_metric(name, value=value)

    def set_estimator(self, estimator):
        self.estimator = estimator

    def set_experiment(self, experiment):
        self.experiment = experiment

    def set_step_name(self, step_name):
        self.step_name = step_name



from rubicon_ml.client import Base


class Parameter(Base):
    """A client parameter.

    A `parameter` is an input to an `experiment` (model run)
    that depends on the type of model being used. It affects
    the model's predictions.

    For example, if you were using a random forest classifier,
    'n_estimators' (the number of trees in the forest) could
    be a parameter.

    A `parameter` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Parameter
        The parameter domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    def __init__(self, domain, config=None):
        super().__init__(domain, config)

    @property
    def id(self):
        """Get the parameter's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the parameter's name."""
        return self._domain.name

    @property
    def value(self):
        """Get the parameter's value."""
        return self._domain.value

    @property
    def description(self):
        """Get the parameter's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the time the parameter was created."""
        return self._domain.created_at



from rubicon_ml.client.asynchronous.config import Config

from rubicon_ml.client.asynchronous.mixin import (  # noqa F401
    ArtifactMixin,
    DataframeMixin,
    TagMixin,
)

from rubicon_ml.client.asynchronous.artifact import Artifact
from rubicon_ml.client.asynchronous.dataframe import Dataframe
from rubicon_ml.client.asynchronous.feature import Feature
from rubicon_ml.client.asynchronous.metric import Metric
from rubicon_ml.client.asynchronous.parameter import Parameter
from rubicon_ml.client.asynchronous.experiment import Experiment
from rubicon_ml.client.asynchronous.project import Project
from rubicon_ml.client.asynchronous.rubicon import Rubicon

__all__ = [
    "Artifact",
    "Config",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "Rubicon",
]



from rubicon_ml.client import Base


class Metric(Base):
    """A client metric.

    A `metric` is a single-value output of an `experiment` that
    helps evaluate the quality of the model's predictions.

    It can be either a 'score' (value to maximize) or
    a 'loss' (value to minimize).

    A `metric` is logged to an `experiment`.

    Parameters
    ----------
    domain : rubicon.domain.Metric
        The metric domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    def __init__(self, domain, config=None):
        super().__init__(domain, config)

    @property
    def id(self):
        """Get the metric's id."""
        return self._domain.id

    @property
    def name(self):
        """Get the metric's name."""
        return self._domain.name

    @property
    def value(self):
        """Get the metric's value."""
        return self._domain.value

    @property
    def directionality(self):
        """Get the metric's directionality."""
        return self._domain.directionality

    @property
    def description(self):
        """Get the metric's description."""
        return self._domain.description

    @property
    def created_at(self):
        """Get the metric's created_at."""
        return self._domain.created_at



import subprocess

import dask.dataframe as dd
import pandas as pd

from rubicon_ml import domain
from rubicon_ml.client import ArtifactMixin, Base, DataframeMixin, Experiment
from rubicon_ml.client.utils.tags import has_tag_requirements


class Project(Base, ArtifactMixin, DataframeMixin):
    """A client project.

    A `project` is a collection of `experiments`,
    `dataframes`, and `artifacts` identified by a unique name.

    Parameters
    ----------
    domain : rubicon.domain.Project
        The project domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    def __init__(self, domain, config=None):
        super().__init__(domain, config)

        self._artifacts = []
        self._dataframes = []
        self._experiments = []

    def _get_branch_name(self):
        """Returns the name of the active branch of the `git` repo
        it is called from.
        """
        command = ["git", "rev-parse", "--abbrev-ref", "HEAD"]
        completed_process = subprocess.run(command, capture_output=True)

        return completed_process.stdout.decode("utf8").replace("\n", "")

    def _get_commit_hash(self):
        """Returns the hash of the last commit to the active branch
        of the `git` repo it is called from.
        """
        command = ["git", "rev-parse", "HEAD"]
        completed_process = subprocess.run(command, capture_output=True)

        return completed_process.stdout.decode("utf8").replace("\n", "")

    def _create_experiment_domain(
        self,
        name,
        description,
        model_name,
        branch_name,
        commit_hash,
        training_metadata,
        tags,
    ):
        """Instantiates and returns an experiment domain object."""
        if self._config.is_auto_git_enabled:
            if branch_name is None:
                branch_name = self._get_branch_name()
            if commit_hash is None:
                commit_hash = self._get_commit_hash()

        if training_metadata is not None:
            training_metadata = domain.utils.TrainingMetadata(training_metadata)

        return domain.Experiment(
            project_name=self._domain.name,
            name=name,
            description=description,
            model_name=model_name,
            branch_name=branch_name,
            commit_hash=commit_hash,
            training_metadata=training_metadata,
            tags=tags,
        )

    def _group_experiments(self, experiments, group_by=None):
        """Groups experiments by `group_by`. Valid options include ["commit_hash"].

        Returns
        -------
        dict
            A dictionary of (group name, DataFrame) key-value pairs.
        """
        GROUP_BY_OPTIONS = ["commit_hash"]
        if group_by is not None and group_by not in GROUP_BY_OPTIONS:
            raise ValueError(f"`group_by` must be one of {GROUP_BY_OPTIONS} or `None`.")

        if group_by is not None:
            grouped_experiments = {}

            if group_by == "commit_hash":
                for experiment in experiments:
                    current_experiments = grouped_experiments.get(experiment.commit_hash, [])
                    current_experiments.append(experiment)
                    grouped_experiments[experiment.commit_hash] = current_experiments
        else:
            grouped_experiments = {None: experiments}

        return grouped_experiments

    def to_dask_df(self, group_by=None):
        """Loads the project's data into dask dataframe(s) sorted by `created_at`.
        This includes the experiment details along with parameters and metrics.

        Parameters
        ----------
        group_by : str or None, optional
            How to group the project's experiments in the returned
            dataframe(s). Valid options include ["commit_hash"].

        Returns
        -------
        dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask dataframe holding the project's
            data. Otherwise a list of dask dataframes holding the project's
            data grouped by `group_by`.
        """
        DEFAULT_COLUMNS = [
            "id",
            "name",
            "description",
            "model_name",
            "commit_hash",
            "tags",
            "created_at",
        ]

        experiments = self.experiments()
        grouped_experiments = self._group_experiments(experiments, group_by=group_by)

        experiment_dfs = {}
        for group, experiments in grouped_experiments.items():
            experiment_records = []
            parameter_names = set()
            metric_names = set()

            for experiment in experiments:
                experiment_record = {
                    "id": experiment.id,
                    "name": experiment.name,
                    "description": experiment.description,
                    "model_name": experiment.model_name,
                    "commit_hash": experiment.commit_hash,
                    "tags": experiment.tags,
                    "created_at": experiment.created_at,
                }

                for parameter in experiment.parameters():
                    experiment_record[f"{parameter.name}"] = parameter.value
                    parameter_names.add(parameter.name)

                for metric in experiment.metrics():
                    experiment_record[f"{metric.name}"] = metric.value
                    metric_names.add(metric.name)

                # TODO - features, artifacts, dataframes represented here?

                experiment_records.append(experiment_record)

            columns = DEFAULT_COLUMNS + list(parameter_names) + list(metric_names)
            df = pd.DataFrame.from_records(experiment_records, columns=columns)

            df = df.sort_values(by=["created_at"], ascending=False).reset_index(drop=True)
            experiment_dfs[group] = dd.from_pandas(df, npartitions=1)

        return experiment_dfs if group_by is not None else list(experiment_dfs.values())[0]

    def log_experiment(
        self,
        name=None,
        description=None,
        model_name=None,
        branch_name=None,
        commit_hash=None,
        training_metadata=None,
        tags=[],
    ):
        """Log a new experiment to this project.

        Parameters
        ----------
        name : str
            The experiment's name.
        description : str, optional
            The experiment's description. Use to provide
            additional context.
        model_name : str, optional
            The experiment's model name. For example, this
            could be the name of the registered model in Model One.
        branch_name : str, optional
            The name of the active branch of the `git` repo this experiment
            is logged from. If omitted and automatic `git` logging is enabled,
            it will be retrieved via `git rev-parse`.
        commit_hash : str, optional
            The hash of the last commit to the active branch of the `git` repo
            this experiment is logged from. If omitted and automatic `git`
            logging is enabled, it will be retrieved via `git rev-parse`.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the experiment's
            training dataset(s).
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your experiments. For example, tags could be used
            to differentiate between the type of model or classifier
            used during the experiment (i.e. `linear regression`
            or `random forest`).

        Returns
        -------
        rubicon.client.Experiment
            The created experiment.
        """
        experiment = self._create_experiment_domain(
            name,
            description,
            model_name,
            branch_name,
            commit_hash,
            training_metadata,
            tags,
        )
        self.repository.create_experiment(experiment)

        return Experiment(experiment, self)

    def experiment(self, id):
        """Get an experiment logged to this project by id.

        Parameters
        ----------
        id : str
            The id of the experiment to get.

        Returns
        -------
        rubicon.client.Experiment
            The experiment logged to this project with id `id`.
        """
        experiment = Experiment(self.repository.get_experiment(self.name, id), self)

        return experiment

    def _filter_experiments(self, experiments, tags, qtype):
        """Filters the provided experiments by `tags` using
        query type `qtype`.
        """
        if len(tags) > 0:
            filtered_experiments = []
            [
                filtered_experiments.append(e)
                for e in experiments
                if has_tag_requirements(e.tags, tags, qtype)
            ]
            self._experiments = filtered_experiments
        else:
            self._experiments = experiments

    def experiments(self, tags=[], qtype="or"):
        """Get the experiments logged to this project.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Experiment
            The experiments previously logged to this project.
        """
        experiments = [Experiment(e, self) for e in self.repository.get_experiments(self.name)]
        self._filter_experiments(experiments, tags, qtype)

        return self._experiments

    def dataframes(self, tags=[], qtype="or", recursive=False):
        """Get the dataframes logged to this project.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.
        recursive : bool, optional
            If true, get the dataframes logged to this project's
            experiments as well. Defaults to false.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        super().dataframes(tags=tags, qtype=qtype)

        if recursive is True:
            for experiment in self.experiments():
                self._dataframes.extend(experiment.dataframes(tags=tags, qtype=qtype))

        return self._dataframes

    @property
    def name(self):
        """Get the project's name."""
        return self._domain.name

    @property
    def id(self):
        """Get the project's id."""
        return self._domain.id

    @property
    def description(self):
        """Get the project's description."""
        return self._domain.description

    @property
    def github_url(self):
        """Get the project's GitHub repository URL."""
        return self._domain.github_url

    @property
    def training_metadata(self):
        """Get the project's training metadata."""
        training_metadata = self._domain.training_metadata.training_metadata

        if len(training_metadata) == 1:
            training_metadata = training_metadata[0]

        return training_metadata

    @property
    def created_at(self):
        """Get the time the project was created."""
        return self._domain.created_at



from rubicon_ml.client import (
    Artifact,
    Dataframe,
    Experiment,
    Feature,
    Metric,
    Parameter,
    Project,
    Rubicon,
)

from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__all__ = [
    "Artifact",
    "Dataframe",
    "Experiment",
    "Feature",
    "Metric",
    "Parameter",
    "Project",
    "Rubicon",
]



import asyncio

import dask.dataframe as dd
import pandas as pd

from rubicon_ml.client import Project as SyncProject
from rubicon_ml.client.asynchronous import ArtifactMixin, DataframeMixin, Experiment
from rubicon_ml.client.utils.tags import has_tag_requirements


class Project(ArtifactMixin, DataframeMixin, SyncProject):
    """An asynchronous client project.

    A `project` is a collection of `experiments`,
    `dataframes`, and `artifacts` identified by a unique name.

    Parameters
    ----------
    domain : rubicon.domain.Project
        The project domain model.
    config : rubicon.client.Config
        The config, which specifies the underlying repository.
    """

    async def _make_experiment_record(self, experiment):
        experiment_record = {
            "id": experiment.id,
            "name": experiment.name,
            "description": experiment.description,
            "model_name": experiment.model_name,
            "commit_hash": experiment.commit_hash,
            "created_at": experiment.created_at,
        }

        parameters_coroutine = experiment.parameters()
        metrics_coroutine = experiment.metrics()

        parameters, metrics = await asyncio.gather(*[parameters_coroutine, metrics_coroutine])

        # tags are being handled differently on purpose to avoid
        # an error when used in asyncio.gather
        experiment_record["tags"] = await experiment.tags

        for parameter in parameters:
            experiment_record[parameter.name] = parameter.value

        for metric in metrics:
            experiment_record[metric.name] = metric.value

        return experiment_record

    async def to_dask_df(self, group_by=None):
        """Overrides `rubicon.client.Project.to_dask_df` to asynchronously
        load the project's data into dask dataframe(s) sorted by `created_at`.

        Parameters
        ----------
        group_by : str or None, optional
            How to group the project's experiments in the returned
            dataframe(s). Valid options include ["commit_hash"].

        Returns
        -------
        dask.DataFrame or list of dask.DataFrame
            If `group_by` is `None`, a dask dataframe holding the project's
            data. Otherwise a list of dask dataframes holding the project's
            data grouped by `group_by`.
        """
        experiments = await self.experiments()
        grouped_experiments = self._group_experiments(experiments, group_by=group_by)

        experiment_dfs = {}
        for group, experiments in grouped_experiments.items():
            experiment_records = await asyncio.gather(
                *[self._make_experiment_record(experiment) for experiment in experiments]
            )

            df = pd.DataFrame.from_records(experiment_records)
            df = df.sort_values(by=["created_at"], ascending=False).reset_index(drop=True)
            experiment_dfs[group] = dd.from_pandas(df, npartitions=1)

        return experiment_dfs if group_by is not None else list(experiment_dfs.values())[0]

    async def log_experiment(
        self,
        name=None,
        description=None,
        model_name=None,
        branch_name=None,
        commit_hash=None,
        training_metadata=None,
        tags=[],
    ):
        """Overrides `rubicon.client.Project.log_experiment` to
        asynchronously log a new experiment to this project.

        Parameters
        ----------
        name : str
            The experiment's name.
        description : str, optional
            The experiment's description. Use to provide
            additional context.
        model_name : str, optional
            The experiment's model name. For example, this
            could be the name of the registered model in Model One.
        branch_name : str, optional
            The name of the active branch of the `git` repo this experiment
            is logged from. If omitted and automatic `git` logging is enabled,
            it will be retrieved via `git rev-parse`.
        commit_hash : str, optional
            The experiment's corresponding git commit hash.
        training_metadata : tuple or list of tuples, optional
            Metadata associated with the experiment's
            training dataset(s).
        tags : list of str, optional
            Values to tag the experiment with. Use tags to organize and
            filter your experiments. For example, tags could be used
            to differentiate between the type of model or classifier
            used during the experiment (i.e. `linear regression`
            or `random forest`).

        Returns
        -------
        rubicon.client.asynchronous.Experiment
            The created experiment.
        """
        experiment = self._create_experiment_domain(
            name,
            description,
            model_name,
            branch_name,
            commit_hash,
            training_metadata,
            tags,
        )
        await self.repository.create_experiment(experiment)

        return Experiment(experiment, self)

    async def experiment(self, id):
        """Overrides `rubicon.client.Project.experiment` to asynchronously
        get an experiment logged to this project by id.

        Parameters
        ----------
        id : str
            The id of the experiment to get.

        Returns
        -------
        rubicon.client.asynchronous.Experiment
            The experiment logged to this project with id `id`.
        """
        experiment = Experiment(await self.repository.get_experiment(self.name, id), self)

        return experiment

    async def _filter_experiments(self, experiments, tags, qtype):
        """Filters the provided experiments by `tags` using
        query type `qtype`.
        """
        if len(tags) > 0:
            filtered_experiments = []
            [
                filtered_experiments.append(e)
                for e in experiments
                if has_tag_requirements(await e.tags, tags, qtype)
            ]
            self._experiments = filtered_experiments
        else:
            self._experiments = experiments

    async def experiments(self, tags=[], qtype="or"):
        """Overrides `rubicon.client.Project.experiments` to asynchronously
        get the experiments logged to this project.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.asynchronous.Experiment
            The experiments previously logged to this project.
        """
        experiments = [
            Experiment(e, self) for e in await self.repository.get_experiments(self.name)
        ]
        await self._filter_experiments(experiments, tags, qtype)

        return self._experiments



import asyncio

import numpy as np
import pandas as pd

from rubicon_ml import Rubicon
from rubicon_ml.client.asynchronous import Rubicon as AsynRubicon


class RubiconModel:
    """The data model for the Rubicon Dashboard.

    Parameters
    ----------
    persistence : str
        The persistence type. Can be one of ["filesystem", "memory"].
    root_dir : str, optional
        Absolute or relative filepath of the root directory holding Rubicon data.
        Use absolute path for best performance. Defaults to the local filesystem.
        Prefix with s3:// to use s3 instead.
    storage_options : dict, optional
        Additional keyword arguments specific to the protocol being chosen. They
        are passed directly to the underlying filesystem class.
    """

    def __init__(self, persistence, root_dir, **storage_options):
        self._rubicon_cls = AsynRubicon if root_dir and root_dir.startswith("s3") else Rubicon
        self._rubicon = self._rubicon_cls(persistence, root_dir, **storage_options)

        self._projects = []
        self._selected_project = None

        self._experiment_table_dfs = {}
        self._experiment_comparison_root_dfs = {}

    def _maybe_run_async(self, rubicon_func, *args, **kwargs):
        """Runs `rubicon_func` asynchronously if logging to S3,
        synchronously otherwise.
        """
        if self._rubicon_cls == AsynRubicon:
            return asyncio.run_coroutine_threadsafe(
                rubicon_func(*args, **kwargs), loop=self._rubicon.repository.filesystem.loop
            ).result()
        else:
            return rubicon_func(*args, **kwargs)

    def get_anchor_options(self, commit_hash):
        """Returns the list of anchor options for the experiments in
        the group with commit hash `commit_hash`.

        Parameters
        ----------
        commit_hash : str
            The commit hash of the group of experiments to get anchor
            options for.

        Returns
        -------
        list of dict
            The anchor options.
        """
        anchors = list(self._experiment_comparison_root_dfs[commit_hash].columns)

        return [{"label": a, "value": a} for a in anchors]

    def get_dimensions(self, commit_hash, selected_experiment_ids, hidden_columns, anchor):
        """Collects and cleans the data necessary to render the Dash parallel
        coordinates plot representing an experiment comparison plot.

        Parameters
        ----------
        commit_hash : str
            The commit hash of the group of experiments to get dimensions for.
        selected_experiment_ids : list of str
            The experiment ID's of the group of experiments to get dimensions
            for. Not all experiments in a group have to be shown on the
            comparison plot.
        hidden_columns : List of columns ids of the columns that are currently
            hidden.
        anchor : str
            The currently selected anchor metric. The anchor metric should be
            represented by the last dimension.

        Returns
        -------
            pandas.Series
                The data corresponding to the selected anchor (to pass to the
                plot as a colorscale).
            list of dict
                The dimensions to render on the plot.
        """
        dimensions = []

        root_df = self.get_experiment_comparison_root_df(commit_hash)
        if hidden_columns is not None:
            # ignore errors since some of the hidden cols could already be dropped, like `id`
            root_df = root_df.drop(columns=hidden_columns, errors="ignore")
        experiment_comparison_df = root_df.loc[selected_experiment_ids]
        experiment_comparison_df = experiment_comparison_df.compute().convert_dtypes()

        for column in experiment_comparison_df.columns:
            clean_column = experiment_comparison_df[column].dropna()

            if isinstance(experiment_comparison_df[column].dtype, pd.StringDtype):
                unique_values = clean_column.unique()
                values = clean_column.map(lambda value: np.where(unique_values == value)[0][0])

                dimension = dict(
                    label=column,
                    ticktext=unique_values,
                    tickvals=list(range(0, len(unique_values))),
                    values=values,
                )
            elif isinstance(experiment_comparison_df[column].dtype, pd.BooleanDtype):
                values = clean_column.astype(int)

                dimension = dict(
                    label=column,
                    ticktext=["False", "True"],
                    tickvals=[0, 1],
                    values=values,
                )
            else:
                values = clean_column

                dimension = dict(label=column, values=values)

            if column == anchor:
                anchor_dimension = dimension
                anchor_data = values
            else:
                dimensions.append(dimension)

        dimensions.append(anchor_dimension)

        return anchor_data, dimensions

    def get_experiment_table_df(self, commit_hash):
        """Returns the experiment table dataframe for the experiments in
        the group with commit hash `commit_hash`.

        Parameters
        ----------
        commit_hash : str
            The commit hash of the group of experiments to get the
            experiment table dataframe for.

        Returns
        -------
        dask.DataFrame
            The dataframe representing the experiment table.
        """
        return self._experiment_table_dfs[commit_hash]

    def get_experiment_comparison_root_df(self, commit_hash):
        """Returns the experiment comparison dataframe for the experiments
        in the group with commit hash `commit_hash`.

        Parameters
        ----------
        commit_hash : str
            The commit hash of the group of experiments to get the
            experiment comparison dataframe for.

        Returns
        -------
        dask.DataFrame
            The dataframe representing the experiment comparison plot.
        """
        return self._experiment_comparison_root_dfs[commit_hash]

    def get_model_names(self, commit_hash):
        """Returns the model name(s) of the experiments in the group with
        commit hash `commit_hash`.
        """
        return list(self._experiment_table_dfs[commit_hash]["model_name"].unique())

    def update_projects(self):
        """Updates the list of available projects and the list
        of dropdown options representing those projects.
        """
        self._projects = self._maybe_run_async(self._rubicon.projects)

    def update_selected_project(self, selected_project_name):
        """Updates the seleted project and loads the relevant data
        for that project.

        Parameters
        ----------
        selected_project_name : str
            The name of the project to set as selected and load data for.
        """
        self._selected_project, *_ = [p for p in self._projects if p.name == selected_project_name]

        grouped_experiment_dfs = self._maybe_run_async(
            self._selected_project.to_dask_df, group_by="commit_hash"
        )

        et_commit_hash_display_length = 7
        et_drop_columns = ["description", "name", "created_at"]
        et_type_castings = {"tags": "str"}

        ec_drop_columns = ["commit_hash", "model_name", "tags"]
        ec_index = "id"

        # no experiments logged to the project yet
        if len(grouped_experiment_dfs.items()) == 0:
            self._experiment_table_dfs = {}

        for commit_hash, df in grouped_experiment_dfs.items():
            experiment_table_df = df.drop(columns=et_drop_columns)
            experiment_table_df = experiment_table_df.astype(et_type_castings)
            experiment_table_df["commit_hash"] = experiment_table_df["commit_hash"].map(
                lambda ch: ch[:et_commit_hash_display_length] if ch is not None else None
            )

            self._experiment_table_dfs[commit_hash] = experiment_table_df

            experiment_comparison_df = experiment_table_df.set_index(ec_index)
            experiment_comparison_df = experiment_comparison_df.drop(columns=ec_drop_columns)

            self._experiment_comparison_root_dfs[commit_hash] = experiment_comparison_df

    @property
    def project_options(self):
        """The available project options.

        Returns
        -------
        list of dict
            The project options.
        """
        return [{"label": p.name, "value": p.name} for p in self._projects]

    @property
    def selected_project(self):
        """The currently selected Rubicon project.

        Returns
        -------
        rubicon.Project or rubicon.client.asynchronous.Project
        """
        return self._selected_project



import intake  # noqa F401

from rubicon_ml.intake_rubicon.experiment import ExperimentSource
from rubicon_ml.intake_rubicon.project import ProjectSource

__all__ = ["ExperimentSource", "ProjectSource"]



def has_tag_requirements(tags, required_tags, qtype):
    """
    Returns True if `tags` meets the requirements based on
    the values of `required_tags` and `qtype`. False otherwise.
    """
    has_tag_requirements = False

    tag_intersection = set(required_tags).intersection(set(tags))
    if qtype == "or":
        if len(tag_intersection) > 0:
            has_tag_requirements = True
    if qtype == "and":
        if len(tag_intersection) == len(required_tags):
            has_tag_requirements = True

    return has_tag_requirements



from rubicon_ml.repository.asynchronous import AsynchronousBaseRepository
from rubicon_ml.repository.utils import json


class S3Repository(AsynchronousBaseRepository):
    """The asynchronous S3 repository uses `asyncio` to
    persist Rubicon data to a remote S3 bucket.

    S3 credentials can be specified via environment variables
    or the credentials file in '~/.aws'.

    Parameters
    ----------
    root_dir : str
        The full S3 path (including 's3://') to persist
        Rubicon data to.
    loop : asyncio.unix_events._UnixSelectorEventLoop, optional
        The event loop the asynchronous calling program is running on.
        It should not be necessary to provide this parameter in
        standard asynchronous operating cases.
    """

    PROTOCOL = "s3"

    async def _connect(self):
        """Asynchronously connect to the underlying S3 persistence layer.

        Note
        ----
        This function must be run before any other that reaches
        out to S3. It is implicitly called by such functions.
        """
        await self.filesystem._connect()

    async def _persist_bytes(self, bytes_data, path):
        """Asynchronously persists the raw bytes `bytes_data`
        to the S3 bucket defined by `path`.
        """
        await self.filesystem._pipe_file(path, bytes_data)

    async def _persist_domain(self, domain, path):
        """Asynchronously persists the Rubicon object `domain`
        to the S3 bucket defined by `path`.
        """
        await self.filesystem._pipe_file(path, json.dumps(domain))



import dash_html_components as html


def make_header_layout():
    """The html layout for the dashboard's header view."""

    return html.Div(
        id="header",
        className="header",
        children=[
            html.Div(id="title", className="header--project", children="rubicon-ml"),
            html.Div(
                id="links",
                className="header--links",
                children=[
                    html.A(
                        className="header--link",
                        href="https://capitalone.github.io/rubicon-ml",
                        children="Docs",
                    ),
                    html.A(
                        className="header--link",
                        href="https://github.com/capitalone/rubicon-ml",
                        children="Github",
                    ),
                ],
            ),
        ],
    )



from dataclasses import dataclass, field
from datetime import datetime

from rubicon_ml.domain.utils import uuid


@dataclass
class Parameter:
    name: str

    id: str = field(default_factory=uuid.uuid4)
    value: object = None
    description: str = None
    created_at: datetime = field(default_factory=datetime.utcnow)



import os

from setuptools import find_packages, setup

import versioneer

pwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(pwd, "README.md"), encoding="utf-8") as readme:
    long_description = readme.read()

install_requires = [
    "click>=7.1",
    "dask[dataframe]>=2.12.0",
    "fsspec>=2021.4.0",
    "intake>=0.5.2",
    "pyarrow>=0.18.0",
    "pyyaml>=5.4.0",
    "s3fs>=0.4",
]

prefect_requires = ["prefect>=0.12.0"]
ui_requires = ["dash>=1.14.0", "dash-bootstrap-components>=0.10.6", "jupyter-dash"]

all_requires = prefect_requires + ui_requires

extras_require = {
    "all": all_requires,
    "prefect": prefect_requires,
    "ui": ui_requires,
}

setup(
    name="rubicon-ml",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Joe Wolfe, Ryan Soley, Diane Lee, Mike McCarty, CapitalOne",
    license="Apache License, Version 2.0",
    description="an ML library for model development and governance",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    url="https://github.com/capitalone/rubicon-ml",
    python_requires=">=3.7",
    install_requires=install_requires,
    extras_require=extras_require,
    entry_points={
        "console_scripts": ["rubicon_ml=rubicon_ml.cli:cli"],
        "intake.drivers": [
            "rubicon_ml_project = rubicon_ml.intake_rubicon.project:ProjectSource",
            "rubicon_ml_experiment = rubicon_ml.intake_rubicon.experiment:ExperimentSource",
        ],
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Documentation",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: Apache Software License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 3 or both.
        "Programming Language :: Python :: 3",
    ],
)



import os
import subprocess
from datetime import datetime

import fsspec

from rubicon_ml import client, domain
from rubicon_ml.client.utils.tags import has_tag_requirements
from rubicon_ml.exceptions import RubiconException


class MultiParentMixin:
    """Adds utils for client objects that can be logged
    to either a `Project` or `Experiment`.
    """

    def _get_parent_identifiers(self):
        """Get the project name and experiment ID (or
        `None`) of this client object's parent(s).
        """
        experiment_id = None

        if isinstance(self, client.Project):
            project_name = self.name
        else:
            project_name = self.project.name
            experiment_id = self.id

        return project_name, experiment_id


class ArtifactMixin(MultiParentMixin):
    """Adds artifact support to a client object."""

    def _validate_data(self, data_bytes, data_file, data_path, name):
        """Raises a `RubiconException` if the data to log as
        an artifact is improperly provided.
        """
        if not any([data_bytes, data_file, data_path]):
            raise RubiconException(
                "One of `data_bytes`, `data_file` or `data_path` must be provided."
            )

        if name is None:
            if data_path is not None:
                name = os.path.basename(data_path)
            else:
                raise RubiconException("`name` must be provided if not using `data_path`.")

        if data_bytes is None:
            if data_file is not None:
                f = data_file
            elif data_path is not None:
                f = fsspec.open(data_path, "rb")

            with f as open_file:
                data_bytes = open_file.read()

        return data_bytes, name

    def log_artifact(
        self, data_bytes=None, data_file=None, data_path=None, name=None, description=None
    ):
        """Log an artifact to this client object.

        Parameters
        ----------
        data_bytes : bytes, optional
            The raw bytes to log as an artifact.
        data_file : TextIOWrapper, optional
            The open file to log as an artifact.
        data_path : str, optional
            The absolute or relative local path or S3 path
            to the data to log as an artifact. S3 paths
            must be prepended with 's3://'.
        name : str, optional
            The name of the artifact file. Required if
            `data_path` is not provided.
        description : str, optional
            A description of the artifact. Use to provide
            additional context.

        Notes
        -----
        Only one of `data_bytes`, `data_file`, and `data_path`
        should be provided. If more than one is given, the order
        of precedence is `data_bytes`, `data_file`, `data_path`.

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Examples
        --------
        >>> # Log with bytes
        >>> experiment.log_artifact(
        ...     data_bytes=b'hello rubicon!', name='bytes_artifact', description="log artifact from bytes"
        ... )

        >>> # Log with file
        >>> with open('some_relevant_file', 'rb') as f:
        >>>     project.log_artifact(
        ...         data_file=f, name='file_artifact', description="log artifact from file"
        ... )

        >>> # Log with file path
        >>> experiment.log_artifact(
        ...     data_path="./path/to/artifact.pkl", description="log artifact from file path"
        ... )
        """
        data_bytes, name = self._validate_data(data_bytes, data_file, data_path, name)

        artifact = domain.Artifact(name=name, description=description, parent_id=self._domain.id)

        project_name, experiment_id = self._get_parent_identifiers()
        self.repository.create_artifact(
            artifact, data_bytes, project_name, experiment_id=experiment_id
        )

        return client.Artifact(artifact, self)

    def _get_environment_bytes(self, export_cmd):
        """Get the working environment as a sequence of bytes.

        Parameters
        ----------
        export_cmd : list of str
            The command to export the environment.

        Returns
        -------
        bytes
            A bytes sequence of the environment.
        """
        try:
            completed_process = subprocess.run(export_cmd, check=True, capture_output=True)
        except subprocess.CalledProcessError as e:
            raise RubiconException(e.stderr)

        return completed_process.stdout

    def log_conda_environment(self, artifact_name=None):
        """Log the conda environment as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported conda environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.

        Notes
        -----
        Relies on running with an active conda environment.
        """
        if artifact_name is None:
            artifact_name = f"environment-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.yml"

        env_bytes = self._get_environment_bytes(["conda", "env", "export"])
        artifact = self.log_artifact(data_bytes=env_bytes, name=artifact_name)

        return artifact

    def log_pip_requirements(self, artifact_name=None):
        """Log the pip requirements as an artifact to this client object.
        Useful for recreating your exact environment at a later date.

        Parameters
        ----------
        artifact_name : str, optional
            The name of the artifact (the exported pip environment).

        Returns
        -------
        rubicon.client.Artifact
            The new artifact.
        """
        if artifact_name is None:
            artifact_name = f"requirements-{datetime.now().strftime('%Y_%m_%d-%I_%M_%S_%p')}.txt"

        requirements_bytes = self._get_environment_bytes(["pip", "freeze"])
        artifact = self.log_artifact(data_bytes=requirements_bytes, name=artifact_name)

        return artifact

    def artifacts(self):
        """Get the artifacts logged to this client object.

        Returns
        -------
        list of rubicon.client.Artifact
            The artifacts previously logged to this client object.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        self._artifacts = [
            client.Artifact(a, self)
            for a in self.repository.get_artifacts_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        return self._artifacts

    def delete_artifacts(self, ids):
        """Delete the artifacts logged to with client object
        with ids `ids`.

        Parameters
        ----------
        ids : list of str
            The ids of the artifacts to delete.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        for artifact_id in ids:
            self.repository.delete_artifact(project_name, artifact_id, experiment_id=experiment_id)


class DataframeMixin(MultiParentMixin):
    """Adds dataframe support to a client object."""

    def log_dataframe(self, df, description=None, tags=[]):
        """Log a dataframe to this client object.

        Parameters
        ----------
        df : pandas.DataFrame or dask.dataframe.DataFrame
            The `dask` or `pandas` dataframe to log.
        description : str, optional
            The dataframe's description. Use to provide
            additional context.
        tags : list of str
            The values to tag the dataframe with.

        Returns
        -------
        rubicon.client.Dataframe
            The new dataframe.
        """
        dataframe = domain.Dataframe(parent_id=self._domain.id, description=description, tags=tags)

        project_name, experiment_id = self._get_parent_identifiers()
        self.repository.create_dataframe(dataframe, df, project_name, experiment_id=experiment_id)

        return client.Dataframe(dataframe, self)

    def _filter_dataframes(self, dataframes, tags, qtype):
        """Filters the provided dataframes by `tags` using
        query type `qtype`.
        """
        if len(tags) > 0:
            filtered_dataframes = []
            [
                filtered_dataframes.append(d)
                for d in dataframes
                if has_tag_requirements(d.tags, tags, qtype)
            ]
            self._dataframes = filtered_dataframes
        else:
            self._dataframes = dataframes

    def dataframes(self, tags=[], qtype="or"):
        """Get the dataframes logged to this client object.

        Parameters
        ----------
        tags : list of str, optional
            The tag values to filter results on.
        qtype : str, optional
            The query type to filter results on. Can be 'or' or
            'and'. Defaults to 'or'.

        Returns
        -------
        list of rubicon.client.Dataframe
            The dataframes previously logged to this client object.
        """
        project_name, experiment_id = self._get_parent_identifiers()
        dataframes = [
            client.Dataframe(d, self)
            for d in self.repository.get_dataframes_metadata(
                project_name, experiment_id=experiment_id
            )
        ]

        self._filter_dataframes(dataframes, tags, qtype)

        return self._dataframes

    def delete_dataframes(self, ids):
        """Delete the dataframes with ids `ids` logged to
        this client object.

        Parameters
        ----------
        ids : list of str
            The ids of the dataframes to delete.
        """
        project_name, experiment_id = self._get_parent_identifiers()

        for dataframe_id in ids:
            self.repository.delete_dataframe(
                project_name, dataframe_id, experiment_id=experiment_id
            )


class TagMixin:
    """Adds tag support to a client object."""

    def _get_taggable_identifiers(self):
        dataframe_id = None

        if isinstance(self, client.Dataframe):
            project_name, experiment_id = self.parent._get_parent_identifiers()
            dataframe_id = self.id
        else:
            project_name, experiment_id = self._get_parent_identifiers()

        return project_name, experiment_id, dataframe_id

    def add_tags(self, tags):
        """Add tags to this client object.

        Parameters
        ----------
        tags : list of str
            The tag values to add.
        """
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()

        self._domain.add_tags(tags)
        self.repository.add_tags(
            project_name, tags, experiment_id=experiment_id, dataframe_id=dataframe_id
        )

    def remove_tags(self, tags):
        """Remove tags from this client object.

        Parameters
        ----------
        tags : list of str
             The tag values to remove.
        """
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()

        self._domain.remove_tags(tags)
        self.repository.remove_tags(
            project_name, tags, experiment_id=experiment_id, dataframe_id=dataframe_id
        )

    def _update_tags(self, tag_data):
        """Add or remove the tags in `tag_data` based on
        their key.
        """
        for tag in tag_data:
            self._domain.add_tags(tag.get("added_tags", []))
            self._domain.remove_tags(tag.get("removed_tags", []))

    @property
    def tags(self):
        """Get this client object's tags."""
        project_name, experiment_id, dataframe_id = self._get_taggable_identifiers()
        tag_data = self.repository.get_tags(
            project_name,
            experiment_id=experiment_id,
            dataframe_id=dataframe_id,
        )

        self._update_tags(tag_data)

        return self._domain.tags



from rubicon_ml.client import Config as SyncConfig
from rubicon_ml.repository.asynchronous import S3Repository


class Config(SyncConfig):
    """Used to configure asynchronous `rubicon` client objects.

    Configuration can be specified (in order of precedence) by:
        1. environment variables 'PERSISTENCE' and 'ROOT_DIR'
        2. arguments to `__init__`

    Parameters
    ----------
    persistence : str, optional
        The persistence type. Can be one of ["filesystem"].
    root_dir : str, optional
        Absolute or relative filepath. Currently, only s3
        paths are supported asynchronously.
    """

    PERSISTENCE_TYPES = ["filesystem"]
    REPOSITORIES = {
        "filesystem-s3": S3Repository,
    }
